import base64
import json
import logging
import os
from abc import ABCMeta, abstractmethod
from typing import Iterable, List

import elastic_transport.client_utils
from configobj import Section
from elastic_transport import NodeConfig, Transport, client_utils
from elastic_transport._transport import TransportApiResponse

from peek.errors import PeekError

_logger = logging.getLogger(__name__)


class NoopDeserializer:
    def __init__(self, delegate):
        self._delegate = delegate

    def dumps(self, *args, **kwargs):
        return self._delegate.dumps(*args, **kwargs)

    def loads(self, s, *args, **kwargs):
        return s

    def get_serializer(self, *args, **kwargs):
        return self._delegate.get_serializer(*args, **kwargs)


class BaseClient(metaclass=ABCMeta):
    @abstractmethod
    def perform_request(self, method, path, payload=None, deserialize_it=False, **kwargs) -> TransportApiResponse:
        pass


class EsClient(BaseClient):
    def __init__(
        self,
        name=None,
        hosts=None,
        cloud_id=None,
        username=None,
        password=None,
        use_ssl=False,
        verify_certs=False,
        assert_hostname=None,
        assert_fingerprint=None,
        ca_certs=None,
        client_cert=None,
        client_key=None,
        api_key=None,
        token=None,
        headers=None,
    ):
        self.name = name
        self.hosts = hosts
        self.cloud_id = cloud_id
        self.auth = f'{username}:{password}' if username and password else None
        self.use_ssl = use_ssl
        self.verify_certs = verify_certs
        self.assert_hostname = assert_hostname
        self.ca_certs = ca_certs
        self.client_cert = client_cert
        self.client_key = client_key
        self.api_key = api_key
        self.token = token
        self.assert_fingerprint = assert_fingerprint
        self.ssl_show_warn = False  # TODO: fill this in as well

        self.headers = headers
        request_headers = {} if self.headers is None else dict(self.headers)

        # The order of authentication scheme is API key, token then basic auth.
        # We add them in reverse order so that later ones overwrite the earlier ones
        # TODO: use basic_auth_to_header utility method from the transport lib
        if self.auth:
            request_headers.update({'Authorization': client_utils.basic_auth_to_header((username, password))})

        if self.token:
            request_headers.update({'Authorization': f'Bearer {self.token}'})

        if self.api_key:
            request_headers.update(
                {'Authorization': 'ApiKey ' + base64.b64encode(':'.join(self.api_key).encode('utf-8')).decode()}
            )

        hosts = []
        if self.hosts:
            for host in self.hosts.split(','):
                if host.startswith('http://') or host.startswith('https://'):
                    hosts.append(host)
                else:
                    hosts.append(f'{"https" if self.use_ssl else "http"}://{host}')

        node_configs = []
        for host in hosts:
            node_config = elastic_transport.client_utils.url_to_node_config(host)
            replacements = {'headers': dict(node_config.headers)}
            replacements['headers'].update(request_headers)
            if node_config.scheme == 'https':
                if self.ca_certs:
                    replacements['ca_certs'] = self.ca_certs
                if self.client_cert:
                    replacements['client_cert'] = self.client_cert
                if self.client_key:
                    replacements['client_key'] = self.client_key
                if self.verify_certs is not None:
                    replacements['verify_certs'] = self.verify_certs
                if self.assert_hostname:
                    replacements['ssl_assert_hostname'] = self.assert_hostname
                if self.assert_fingerprint:
                    replacements['ssl_assert_fingerprint'] = self.assert_fingerprint
                if self.ssl_show_warn is not None:
                    replacements['ssl_show_warn'] = self.ssl_show_warn

                # By default verify_certs is false. If client cert and key are explicitly passed,
                # verify_certs must be true as well, otherwise Python won't send the client materials at all
                if 'client_cert' in replacements or 'client_key' in replacements:
                    replacements['verify_certs'] = True

            node_configs.append(node_config.replace(**replacements))

        if self.cloud_id:
            cloud_id = elastic_transport.client_utils.parse_cloud_id(self.cloud_id)
            node_config = NodeConfig(
                scheme='https', host=cloud_id.es_address[0], port=cloud_id.es_address[1], http_compress=True
            )
            node_configs.append(node_config.replace(headers=request_headers))

        if not node_configs:
            raise ValueError('no node configurations found')

        # Timeout is handled at perform_request
        self.transport = Transport(node_configs, max_retries=0, retry_on_timeout=False)

    def perform_request(self, method, path, payload=None, deserialize_it=False, headers=None, **kwargs):
        _logger.debug(f'Performing request: {method!r}, {path!r}, {payload!r}')
        serializers = self.transport.serializers
        try:
            if not deserialize_it:
                # Avoid deserializing the response since we parse it with the main loop for syntax highlighting
                self.transport.serializers = NoopDeserializer(serializers)

            http_headers = elastic_transport.HttpHeaders(headers)

            if payload is not None and 'content-type' not in http_headers:
                http_headers['content-type'] = 'application/json'

            response = self.transport.perform_request(
                method, path, body=payload, request_timeout=None, headers=http_headers, **kwargs
            )

            if deserialize_it:
                return response
            else:
                return TransportApiResponse(response.meta, response.body.decode('utf-8'))
        finally:
            if not deserialize_it:
                self.transport.serializers = serializers

    def info(self):
        if self.api_key:
            auth = f'ApiKey {self.api_key[0][:10]}...'
        elif self.token:
            auth = f'Token {self.token[:10]}...'
        elif self.auth:
            auth = f'Username {self.auth.split(":")[0]}'
        else:
            auth = None

        if self.headers:
            headers = list(self.headers.keys())
        else:
            headers = None

        return {
            'name': self.name,
            'hosts': self.hosts,
            'cloud_id': self.cloud_id,
            'auth': auth,
            'use_ssl': self.use_ssl,
            'verify_certs': self.verify_certs,
            'ca_certs': self.ca_certs,
            'client_cert': self.client_cert,
            'client_key': self.client_key,
            'headers': headers,
        }

    def to_dict(self):
        return {
            'name': self.name,
            'hosts': self.hosts,
            'cloud_id': self.cloud_id,
            'username': self.auth.split(':')[0] if self.auth else None,
            'password': None,  # do not persist password
            'use_ssl': self.use_ssl,
            'verify_certs': self.verify_certs,
            'assert_hostname': self.assert_hostname,
            'assert_fingerprint': self.assert_fingerprint,
            'ca_certs': self.ca_certs,
            'client_cert': self.client_cert,
            'client_key': self.client_key,
            'api_key': ':'.join(self.api_key) if self.api_key else None,
            'token': self.token,
            'headers': self.headers,
        }

    @staticmethod
    def from_dict(app, d):
        return connect(app, **d)

    def __str__(self):
        if self.name:
            return f'{self.name}'

        hosts = []
        if self.hosts:
            for host in self.hosts.split(','):
                if host.startswith('https://') or host.startswith('http://'):
                    hosts.append(host)
                else:
                    hosts.append(('https://' if self.use_ssl else 'http://') + host)
        else:
            hosts.append(self.cloud_id.split(':')[0] + ' [Cloud]')

        hosts = ','.join(hosts)
        if self.api_key:
            return f'K-{self.api_key[0][:10]} @ {hosts}'
        elif self.token:
            return f'T-{self.token[:10]} @ {hosts}'
        elif self.auth:
            username = self.auth.split(':')[0]
            return f'{username} @ {hosts}'
        else:
            return f'{hosts}'


class RefreshingEsClient(BaseClient):
    def __init__(self, parent: EsClient, username, access_token, refresh_token, expires_in, name=None):
        self.parent = parent
        self.username = username
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expires_in = expires_in
        self.name = name
        self.delegate = self._build_delegate()

    def __getattr__(self, item):
        return getattr(self.delegate, item)

    def perform_request(self, method, path, payload=None, deserialize_it=False, **kwargs):
        response = self.delegate.perform_request(method, path, payload, deserialize_it, **kwargs)
        if response.meta.status == 401:
            body = self.parent.perform_request(
                'POST',
                '/_security/oauth2/token',
                json.dumps(
                    {
                        'grant_type': 'refresh_token',
                        'refresh_token': self.refresh_token,
                    }
                ),
                deserialize_it=True,
            ).body
            self.access_token = body['access_token']
            self.refresh_token = body['refresh_token']
            self.expires_in = body['expires_in']
            self.delegate = self._build_delegate()
            return self.perform_request(method, path, payload, deserialize_it=deserialize_it, **kwargs)
        else:
            return response

    def info(self):
        info = self.delegate.info()
        info['username'] = self.username
        return info

    def to_dict(self):
        return {
            'name': self.name,
            'username': self.username,
            'access_token': self.access_token,
            'refresh_token': self.refresh_token,
            'expires_in': self.expires_in,
            'parent': self.parent,
        }

    @staticmethod
    def from_dict(d):
        return RefreshingEsClient(**d)

    def __str__(self):
        if self.name:
            return f'{self.name}'
        else:
            return f'{self.username} @ {self.delegate}'

    def _build_delegate(self):
        return EsClient(
            name=self.name,
            hosts=self.parent.hosts,
            cloud_id=self.parent.cloud_id,
            use_ssl=self.parent.use_ssl,
            verify_certs=self.parent.verify_certs,
            assert_hostname=self.parent.assert_hostname,
            assert_fingerprint=self.parent.assert_fingerprint,
            ca_certs=self.parent.ca_certs,
            client_cert=self.parent.client_cert,
            client_key=self.parent.client_key,
            headers=self.parent.headers,
            token=self.access_token,
        )


class DelegatingListener:
    def __init__(self, on_add=None, on_set=None, on_remove=None):
        self._on_add = on_add
        self._on_set = on_set
        self._on_remove = on_remove

    def on_add(self, m):
        return self._on_add(m) if self._on_add is not None else True

    def on_set(self, m):
        return self._on_set(m) if self._on_set is not None else True

    def on_remove(self, m, c):
        return self._on_remove(m, c) if self._on_remove is not None else True


class EsClientManager:
    def __init__(self, listeners: Iterable[DelegatingListener] = ()):
        self._clients: List[EsClient] = []
        self._index_current = None
        self.listeners = listeners

    def add(self, client):
        self._clients.append(client)
        self._index_current = len(self._clients) - 1
        # TODO: maintain size
        for listener in self.listeners:
            if listener.on_add(self) is False:
                break

    @property
    def index_current(self):
        return self._index_current

    @property
    def current(self):
        if self._index_current is None:
            raise PeekError('No ES client is configured')
        if self._index_current < 0 or self._index_current >= len(self._clients):
            raise PeekError(f'Attempt to get ES client at invalid index [{self._index_current}]')
        return self._clients[self._index_current]

    def set_current(self, x):
        if isinstance(x, str):
            name = x
            self._index_current = self._clients.index(self.get_client(name))
        elif isinstance(x, int):
            if x < 0 or x >= len(self._clients):
                raise PeekError(f'Attempt to set ES client at invalid index {x!r}')
            self._index_current = x
        else:
            raise ValueError(f'Connection must be specified by either name or index, got {x!r}')
        for listener in self.listeners:
            if listener.on_set(self) is False:
                break

    def clients(self):
        return self._clients

    def get_client(self, x=None):
        if x is None:
            return self.current
        elif isinstance(x, str):
            name = x
            if not name:
                raise ValueError('Name cannot be empty')
            for c in self._clients:
                if c.name == name:
                    return c
            else:
                raise ValueError(f'No client with name: {name!r}')
        elif isinstance(x, int):
            if x < 0 or x >= len(self._clients):
                raise PeekError(f'Attempt to set ES client at invalid index [{x}]')
            return self._clients[x]
        else:
            raise ValueError(f'Connection must be specified by either name or index, got {x!r}')

    def remove_client(self, x=None):
        if x is None:
            self.remove_client(self._index_current)
        elif isinstance(x, str):
            idx = self._clients.index(self.get_client(x))
            self.remove_client(idx)
        elif isinstance(x, int):
            if x < 0 or x >= len(self._clients):
                raise PeekError(f'Attempt to remove ES client at invalid index [{x}]')
            removed = self._clients.pop(x)
            if not self._clients:
                self._index_current = None
                return
            if x < self._index_current:
                self._index_current -= 1
            elif x == self._index_current:
                self._index_current = 0
            for listener in self.listeners:
                if listener.on_remove(self, removed) is False:
                    break
        else:
            raise ValueError(f'Connection must be specified by either name or index, got {x!r}')

    def keep_client(self, x=None):
        """
        Keep the specified client, remove everything else
        """
        keep = self.get_client(x)
        self._index_current = 0
        self._clients = [keep]

    def move_current_to(self, idx: int):
        if not isinstance(idx, int):
            raise PeekError(f'Index must be integer, got {idx!r}')
        if 0 <= idx < len(self._clients):
            if idx == self._index_current:
                return
            self._clients.insert(idx, self._clients.pop(self._index_current))
            self._index_current = idx
        else:
            raise PeekError(f'Attempt to move ES client to an invalid index: {idx!r}')

    def to_dict(self):
        result = {
            '_index_current': self._index_current,
            '_clients': [],
        }
        for client in self._clients:
            d = client.to_dict()
            if isinstance(client, RefreshingEsClient):
                try:
                    d['parent'] = self._clients.index(d['parent'])
                except ValueError:
                    d['parent'] = d['parent'].to_dict()
            result['_clients'].append(d)
        return result

    @staticmethod
    def from_dict(app, d):
        m = EsClientManager()
        _clients = []
        for c in d['_clients']:
            if 'parent' in c:
                c = dict(c)  # avoid mutating argument
                if isinstance(c['parent'], int):
                    # Based on how clients are arranged, it is guaranteed that parent client must have
                    # an index less than the refreshing client
                    c['parent'] = _clients[c['parent']]
                else:
                    c['parent'] = EsClient.from_dict(app, c['parent'])
                client = RefreshingEsClient.from_dict(c)
            else:
                client = EsClient.from_dict(app, c)
            _clients.append(client)

        m._clients = _clients
        m._index_current = d['_index_current']
        return m

    def __str__(self):
        lines = []
        for i, client in enumerate(self.clients()):
            prefix = '*' if client == self.current else ' '
            index = f'[{i}]'
            lines.append(f'{prefix} {index:>4} {client}')
        return '\n'.join(lines)


DEFAULT_OPTIONS = {
    'name': None,
    'hosts': 'localhost:9200',
    'cloud_id': None,
    'username': None,
    'password': None,
    'api_key': None,
    'token': None,
    'use_ssl': False,
    'verify_certs': False,
    'assert_hostname': None,
    'assert_fingerprint': None,
    'ca_certs': None,
    'client_cert': None,
    'client_key': None,
    'headers': None,
    'force_prompt': False,
    'no_prompt': False,
}


def connect(app, **options):
    final_options = dict(DEFAULT_OPTIONS)

    if isinstance(app.config.get('connection'), Section):
        final_options.update({k: v for k, v in app.config.get('connection').dict().items() if v})

    final_options.update(_maybe_configure_smart_connect(app, options))

    if final_options['cloud_id'] is not None:
        final_options['hosts'] = None

    if final_options['api_key']:
        return _connect_api_key(app, **final_options)
    elif final_options['token']:
        return _connect_token(app, **final_options)
    else:
        return _connect_userpass(app, **final_options)


def _maybe_configure_smart_connect(app, options: dict):
    # Override with provided options, including null values since it could be intentional
    smart_options = {k: v for k, v in options.items()}
    # Attempt to configure smart connect based on the last HTTP response
    if len(options) == 0:
        try:
            last_response = app.vm.get_value('_')
            if 'id' in last_response and 'api_key' in last_response:
                _maybe_copy_current_client_options(app, smart_options)
                smart_options['api_key'] = f'{last_response["id"]}:{last_response["api_key"]}'
            elif 'access_token' in last_response:
                _maybe_copy_current_client_options(app, smart_options)
                smart_options['token'] = last_response['access_token']
            elif 'token' in last_response and 'value' in last_response['token']:
                _maybe_copy_current_client_options(app, smart_options)
                smart_options['token'] = last_response['token']['value']
            else:
                last_request = app.vm.get_value('__')
                urlpath = last_request['path']
                if last_request['method'] in ('POST', 'PUT') and (
                    urlpath.startswith('/_security/user/') or urlpath.startswith('/_xpack/security/user/')
                ):
                    if urlpath.endswith('/_password'):
                        endpos = len(urlpath) - len('/_password')
                    else:
                        endpos = len(urlpath)
                    if urlpath.startswith('/_security/user/'):
                        username = urlpath[len('/_security/user/') :endpos]
                    else:
                        username = urlpath[len('/_xpack/security/user/') :endpos]
                    payload = json.loads(last_request['payload'])
                    password = payload.get('password', None)
                    _maybe_copy_current_client_options(app, smart_options)
                    smart_options['username'] = username
                    smart_options['password'] = password

        except NameError:  # if _ does not exist, simply ignore and proceed
            pass

    return smart_options


def _maybe_copy_current_client_options(app, options: dict):
    if hasattr(app, 'es_client_manager') and len(app.es_client_manager.clients()) > 0:
        current_es_client = app.es_client_manager.current
        if current_es_client.cloud_id is not None:
            options['cloud_id'] = current_es_client.cloud_id
        else:
            options['hosts'] = current_es_client.hosts

        options['use_ssl'] = current_es_client.use_ssl
        options['verify_certs'] = current_es_client.verify_certs
        options['assert_hostname'] = current_es_client.assert_hostname
        options['assert_fingerprint'] = current_es_client.assert_fingerprint
        options['ca_certs'] = current_es_client.ca_certs
        options['client_cert'] = current_es_client.client_cert
        options['client_key'] = current_es_client.client_key
        # not copy the headers


def _connect_userpass(app, **options):
    username = options.get('username')
    password = options.get('password')
    if options['hosts']:
        service_name = f'peek/{options["hosts"]}/userpass'
    else:
        service_name = f'peek/{options["cloud_id"]}/userpass'

    if not username and password:
        raise PeekError('Username is required for userpass authentication')

    if options['force_prompt']:
        password = app.input(message='Please enter password: ', is_secret=True)

    if username and not password:
        password = os.environ.get('PEEK_PASSWORD', None)
        if not password:
            if app.config.as_bool('use_keyring'):
                password = _keyring(service_name, username)
                if not password:
                    if options['no_prompt']:
                        raise PeekError('Password is not found and password prompt is disabled')
                    password = app.input(message='Please enter password: ', is_secret=True)

            else:
                if options['no_prompt']:
                    raise PeekError('Password is not found and password prompt is disabled')
                password = app.input(message='Please enter password: ', is_secret=True)

    if username and password and app.config.as_bool('use_keyring'):
        _keyring(service_name, username, password)

    return EsClient(
        name=options['name'],
        hosts=options['hosts'],
        cloud_id=options['cloud_id'],
        username=username,
        password=password,
        use_ssl=options['use_ssl'],
        verify_certs=options['verify_certs'],
        assert_hostname=options['assert_hostname'],
        assert_fingerprint=options['assert_fingerprint'],
        ca_certs=options['ca_certs'],
        client_cert=options['client_cert'],
        client_key=options['client_key'],
        headers=options['headers'],
    )


def _connect_api_key(app, **options):
    _logger.debug('Connecting with API key')
    api_key = options['api_key']
    if ':' not in api_key:
        api_key = base64.decodebytes(api_key.encode()).decode()

    if ':' not in api_key:
        raise ValueError('invalid api key credential format')

    return EsClient(
        name=options['name'],
        hosts=options['hosts'],
        cloud_id=options['cloud_id'],
        api_key=api_key.split(':'),
        use_ssl=options['use_ssl'],
        verify_certs=options['verify_certs'],
        assert_hostname=options['assert_hostname'],
        assert_fingerprint=options['assert_fingerprint'],
        ca_certs=options['ca_certs'],
        client_cert=options['client_cert'],
        client_key=options['client_key'],
        headers=options['headers'],
    )


def _connect_token(app, **options):
    _logger.debug('Connecting with token')
    return EsClient(
        name=options['name'],
        hosts=options['hosts'],
        cloud_id=options['cloud_id'],
        token=options['token'],
        use_ssl=options['use_ssl'],
        verify_certs=options['verify_certs'],
        assert_hostname=options['assert_hostname'],
        assert_fingerprint=options['assert_fingerprint'],
        ca_certs=options['ca_certs'],
        client_cert=options['client_cert'],
        client_key=options['client_key'],
        headers=options['headers'],
    )


KEYRING = None


def _keyring(service_name, key, value=None):
    """
    When a None value is passed in, means getting value out of the keyring
    """
    global KEYRING
    if KEYRING is None:
        import importlib

        try:
            KEYRING = importlib.import_module("keyring")
        except ModuleNotFoundError as e:
            _logger.warning("import keyring failed: %r.", e)

    if KEYRING is not None:
        if value is None:
            return KEYRING.get_password(service_name, key)
        else:
            KEYRING.set_password(service_name, key, value)


class ConnectFunc:
    def __call__(self, app, **options):
        test_connection = app.config.as_bool('test_connection')
        opt = options.pop('test', None)
        if opt is not None:
            test_connection = opt
        es_client = connect(app, **options)
        if test_connection:
            try:
                app.display.info(es_client.perform_request('GET', '/_security/_authenticate').body)
            except Exception as e:
                app.display.error(e)
                return
        app.es_client_manager.add(es_client)
        return str(app.es_client_manager)

    @property
    def options(self):
        _options = dict(DEFAULT_OPTIONS)
        _options['test'] = None
        return _options

    @property
    def description(self):
        return 'Connect to an Elasticsearch cluster'



import json
import logging
from typing import Any, Tuple

from prompt_toolkit.buffer import CompletionState
from prompt_toolkit.completion import Completion

_logger = logging.getLogger(__name__)


class PayloadKeyCompletion(Completion):
    def __init__(self, text: str, value, *args, **kwargs):
        super().__init__(text, *args, **kwargs)
        self.value = value


def proxy_new_text_and_position(self: CompletionState) -> Tuple[str, int]:
    if self.complete_index is None:
        return self.original_new_text_and_position()

    c = self.completions[self.complete_index]
    if not isinstance(c, PayloadKeyCompletion):
        return self.original_new_text_and_position()

    # check whether original text and cursor is something like "CURSOR"
    # Optionally check whether new text is followed by a quote and
    # remove the quote
    original_text_before_cursor = self.original_document.text_before_cursor
    original_text_after_cursor = self.original_document.text_after_cursor

    _logger.debug(f'text_before_cursor: {original_text_before_cursor!r}')
    _logger.debug(f'start_position: {c.start_position!r}')
    end_position = len(original_text_before_cursor) if c.start_position == 0 else c.start_position
    quote = original_text_before_cursor[end_position - 3 : end_position]
    _logger.debug(f'quote: {quote!r}')
    if quote not in ('"""', "'''"):
        quote = quote[-1]

    idx_quote_end = original_text_after_cursor.find(quote)
    if idx_quote_end == -1:
        idx_quote_end = original_text_after_cursor.find('\n')
        if idx_quote_end == -1:
            idx_quote_end = len(original_text_after_cursor) - 1
    idx_quote_end += 1

    before = original_text_before_cursor[: c.start_position - len(quote)]

    current_indent = len(self.original_document.current_line) - len(self.original_document.current_line.lstrip())

    key_fill = f'{json.dumps(c.text)}: '
    value_fill = get_value_fill(c.value, current_indent)

    before = before + key_fill
    new_text = before + value_fill + original_text_after_cursor[idx_quote_end:]

    seen_left_curly = False
    seen_quote = 0
    for idx, c in enumerate(value_fill):
        if c == '{':
            seen_left_curly = True

        if c in (']', ')', '}'):
            new_cursor_position = len(before) + idx
            break
        elif c in ('"', "'"):
            if not seen_left_curly or (seen_left_curly and seen_quote >= 2):
                new_cursor_position = len(before) + idx + 1
                break
            else:
                seen_quote += 1
        elif c == ',':
            new_cursor_position = len(before) + idx - 1
            break
    else:
        new_cursor_position = len(before)

    return new_text, new_cursor_position


def get_value_fill(value: Any, current_indent: int):
    if isinstance(value, dict):
        if '__template' in value:
            value_fill = serialise_and_indent_json(value['__template'], current_indent)
        elif '__one_of' in value:
            return get_value_fill(value['__one_of'][0], current_indent)
        else:
            value_fill = '{}'
    elif isinstance(value, list):
        if len(value) > 0 and isinstance(value[0], dict):
            value_fill = serialise_and_indent_json([{}], current_indent)
        else:
            value_fill = '[]'
    else:
        value_fill = json.dumps(value)

    value_fill += ', '

    return value_fill


def serialise_and_indent_json(data, current_indent):
    indent = current_indent * ' '
    res = json.dumps(data, indent=2)
    return res.replace('\n', f'\n{indent}')


def monkey_patch_completion_state():
    if CompletionState.new_text_and_position != proxy_new_text_and_position:
        CompletionState.original_new_text_and_position = CompletionState.new_text_and_position
        CompletionState.new_text_and_position = proxy_new_text_and_position



import re
from typing import Iterable

from pygments.lexer import RegexLexer, bygroups, default, include, words
from pygments.style import Style
from pygments.token import (
    Comment,
    Error,
    Generic,
    Keyword,
    Literal,
    Name,
    Number,
    Operator,
    Punctuation,
    String,
    Text,
    Whitespace,
)

from peek.common import HTTP_METHODS, PeekToken
from peek.errors import PeekSyntaxError

Percent = Punctuation.Percent
CurlyLeft = Punctuation.Curly.Left
CurlyRight = Punctuation.Curly.Right
BracketLeft = Punctuation.Bracket.Left
BracketRight = Punctuation.Bracket.Right
ParenLeft = Punctuation.Paren.Left
ParenRight = Punctuation.Paren.Right
Comma = Punctuation.Comma
Colon = Punctuation.Colon
At = Punctuation.At
Heading = Generic.Heading
TripleD = String.TripleD
TripleS = String.TripleS
Assign = Operator.Assign
BinOp = Operator.BinOp
UnaryOp = Operator.UnaryOp
BlankLine = Whitespace.BlankLine
FuncName = Name.Variable
HttpMethod = Keyword.HttpMethod
Let = Keyword.Let
For = Keyword.For
In = Keyword.In
ShellOut = Punctuation.Bang
DictKey = String.Symbol
OptionName = Name.Symbol

TipsMinor = Generic.TipsMinor
EOF = Whitespace.EOF

VARIABLE_PATTERN = r'\b[_a-zA-Z]+[_a-zA-Z0-9]*\b'


class PeekStyle(Style):
    default_style = ''
    # null: #445
    styles = {
        Keyword: '#03F147',
        Literal: '#FFFCF9',
        DictKey: '#bff',  # '#28b',
        String: '#395',
        Name.Builtin: '#77f',
        FuncName: '#77f',
        Number: '#07a',
        Heading: '#F6D845',
        TipsMinor: '#4C4447',
        Error: 'bg:#a40000',
    }


def dqs(ttype):
    return [
        (r'"', ttype, '#pop'),  # end of string
        (r'\\.', ttype),  # process escapes separately
        # Inner string, not allow multiline, not processing end of string,
        # also don't consume a single backslash on its own since it is processed above
        # in combination with the char it applies to.
        (r'[^\\\n"]+', ttype),
    ]


def sqs(ttype):
    return [
        (r"'", ttype, '#pop'),
        (r"\\.", ttype),
        (r'[^\\\n\']+', ttype),
    ]


W = r'[ \t\r\f\v]'


# Whitespace should be consumed in parent state, so that child state can simply pop on default no matching
# If there is a signature pattern to signal a #pop, use it and it's ok to consume whitespace. Otherwise,
# default to pop and do not allow consuming whitespace standalone.


class PeekLexer(RegexLexer):
    name = 'ES'
    aliases = ['es']
    filenames = ['*.es']

    tokens = {
        'root': [
            (r'(!)(.*)', bygroups(ShellOut, Literal)),
            (r'//.*', Comment.Single),
            (
                r'(?i)(' + '|'.join(m.upper() for m in HTTP_METHODS) + r')\b(' + W + '*)',
                bygroups(HttpMethod, Whitespace),
                'api_path',
            ),
            # TODO: more keywords
            (r'(let)\b(' + W + '*)', bygroups(Let, Whitespace), 'let_args'),
            (r'(for)\b(' + W + '*)', bygroups(For, Whitespace), 'for_name'),
            (VARIABLE_PATTERN, FuncName, 'func_stmt_args'),
            (r'\s+', Whitespace),
        ],
        'stmts': [include('root'), default('#pop')],
        'let_args': [
            (r'\n', BlankLine, '#pop'),
            (r'//.*', Comment.Single),
            (W + r'*(?=\S)', Whitespace, ('assign_rhs', 'value')),
        ],
        'for_name': [
            (VARIABLE_PATTERN, Name, ('#pop', 'for_in')),
        ],
        'for_in': [
            (W + r'+', Whitespace),
            (r'(in)\b(' + W + r'*)', bygroups(In, Whitespace), ('#pop', 'for_body_start', 'value')),
        ],
        'for_body_start': [
            (r'(' + W + r'*)(\{)', bygroups(Whitespace, CurlyLeft), ('#pop', 'for_body_stop', 'stmts')),
        ],
        'for_body_stop': [
            (r'(\s*)(\})', bygroups(Whitespace, CurlyRight), '#pop'),
        ],
        'api_path': [
            (r'(\s*)(?=\()', Whitespace, ('#pop', 'api_options', 'group')),
            (r'\S+', Literal, ('#pop', 'api_options')),
        ],
        'api_options': [
            (r'\n', Whitespace, ('#pop', 'payload')),
            (r'//.*', Comment.Single),
            (VARIABLE_PATTERN, OptionName, 'assign_rhs'),
            (W + r'+', Whitespace),
            # default(('#pop', 'payload')),  # this is to make the newline optional
        ],
        'assign_rhs': [
            (r'(' + W + r'*)(=)(' + W + r'*)', bygroups(Whitespace, Assign, Whitespace), ('#pop', 'value')),
        ],
        'payload': [
            (r'(' + W + '*)' + r'(//.*)(\n)', bygroups(Whitespace, Comment.Single, Whitespace)),
            (r'(' + W + r'*)(?={)', Whitespace, ('#pop', 'payload_cont', 'dict')),
            (r'(' + W + r'*)(@)', bygroups(Whitespace, At), ('#pop', 'payload_file')),
            default('#pop'),  # payload has default pop because it is optional and next statement may begin here
        ],
        'payload_file': [
            (r'[^\n]+', Literal, '#pop'),
        ],
        'payload_cont': [
            (r'(\n' + W + r'*)(//.*)', bygroups(Whitespace, Comment.Single)),
            (r'(\n' + W + r'*)(?={)', Whitespace, 'dict'),
            default('#pop'),  # similar to payload, payload_cont can stop at anytime and next statement may begin
        ],
        'func_stmt_args': [
            (r'\n', BlankLine, '#pop'),
            (r'//.*', Comment.Single),
            (VARIABLE_PATTERN + r'(?=' + W + '*=)', Name, 'func_option'),  # kv pair
            (W + r'+', Whitespace),
            (r',', Whitespace),
            # Instead of include, we should default to get into value because value pop itself out of the stack
            # one it is consumed. If included it, it will pop out the func_stmt_args stage, which is incorrect.
            # func_stmt_args needs has explicit pop by newline.
            default('value'),
        ],
        'func_option': [
            (r'(' + W + r'*)(=)(' + W + r'*)', bygroups(Whitespace, Assign, Whitespace), ('#pop', 'value')),
            default('#pop'),  # func option has default pop because func has positional arg
        ],
        'dict': [
            (r'({)(\s*)', bygroups(CurlyLeft, Whitespace), ('colon', 'dict_key')),
            (r'//.*', Comment.Single),
            (r'\s+', Whitespace),
            (r'}', CurlyRight, '#pop'),
            (r'(,)(\s*)', bygroups(Comma, Whitespace), ('colon', 'dict_key')),
        ],
        # dict_key should not have default pop because of similar reason to value
        'dict_key': [
            (r'(\s*)(//.*)(\s*)', bygroups(Whitespace, Comment.Single, Whitespace)),
            (r'(\s*)(?=})', Whitespace, '#pop:2'),  # special handle for empty dict and extra comma
            # dict_key does not support triple quotes by itself, but can be used if it is part of an expression
            (r'"(?!"")', DictKey, ('#pop', 'operators', 'dqs_key')),
            (r"'(?!'')", DictKey, ('#pop', 'operators', 'sqs_key')),
            include('value'),
        ],
        'colon': [
            (r'(\s*)(//.*)(\s*)', bygroups(Whitespace, Comment.Single, Whitespace)),
            (r'(\s*)(:)(\s*)', bygroups(Whitespace, Colon, Whitespace), ('#pop', 'dict_value')),
        ],
        'dict_value': [
            (r'(\s*)(//.*)(\s*)', bygroups(Whitespace, Comment.Single, Whitespace)),
            include('value'),
        ],
        'array': [
            (r'(\[)(\s*)', bygroups(BracketLeft, Whitespace), 'array_value'),
            (r'//.*', Comment.Single),
            (r'\s+', Whitespace),
            (r']', BracketRight, '#pop'),
            (r'(,)(\s*)', bygroups(Comma, Whitespace), 'array_value'),
        ],
        'array_value': [
            (r'(\s*)(//.*)(\s*)', bygroups(Whitespace, Comment.Single, Whitespace)),
            (r'(\s*)(?=\])', Whitespace, '#pop'),  # special handle for empty array and extra comma
            include('value'),
        ],
        'group': [
            (r'(\()(\s*)', bygroups(ParenLeft, Whitespace), 'value'),
            (r'//.*', Comment.Single),
            (r'\s+', Whitespace),
            (r'\)', ParenRight, '#pop'),
        ],
        'group_value': [
            (r'(\s*)(//.*)(\s*)', bygroups(Whitespace, Comment.Single, Whitespace)),
            (r'(\s*)(?=\))', Whitespace, '#pop'),  # special handle for empty array and extra comma
            include('value'),
        ],
        # Single value should NOT default pop because when the lexer gets here, it should
        # expect exact one value to consume, anything else should be handled by the upstream.
        '_value_common': [
            (r'(' + W + r'*)([-+]?)(?=\()', bygroups(Whitespace, UnaryOp), ('#pop', 'operators', 'group')),
            (r'(' + W + r'*)(?={)', Whitespace, ('#pop', 'operators', 'dict')),
            (r'(' + W + r'*)(?=\[)', Whitespace, ('#pop', 'operators', 'array')),
            (r'"""', TripleD, ('#pop', 'operators', 'tdqs')),
            (r"'''", TripleS, ('#pop', 'operators', 'tsqs')),
            (r'"', String.Double, ('#pop', 'operators', 'dqs')),
            (r"'", String.Single, ('#pop', 'operators', 'sqs')),
            (words(('true', 'false', 'null'), suffix=r'\b'), Name.Builtin, ('#pop', 'operators')),
            # (r'([-+]?)(' + VARIABLE_PATTERN + r')(' + W + r'*)(?=\()',
            #  bygroups(UnaryOp, Name, Whitespace), ('#pop', 'func_expr')),
            (r'([-+]?)(' + VARIABLE_PATTERN + r')', bygroups(UnaryOp, Name), ('#pop', 'operators')),
            (r'@', At, ('#pop', 'operators', 'symbol')),
        ],
        'symbol': [
            (VARIABLE_PATTERN, Literal, '#pop'),
        ],
        'value': [
            include('_value_common'),
            include('numbers'),
        ],
        'dotable_value': [
            include('_value_common'),
            include('integers'),
        ],
        'func_expr': [
            (r'//.*', Comment.Single),
            (W + r'+', Whitespace),
            (r'(\()(\s*)', bygroups(ParenLeft, Whitespace), 'func_expr_args'),
            (r'\)', ParenRight, '#pop'),
        ],
        'func_expr_args': [
            (r'//.*', Comment.Single),
            (r'(\s*)(?=\))', Whitespace, '#pop'),
            (VARIABLE_PATTERN + r'(?=' + W + '*=)', Name, 'func_option'),  # kv pair
            (r'\s+', Whitespace),
            (r',', Whitespace),
            default('value'),
        ],
        'operators': [  # operator cannot be on the new line
            (W + r'+', Whitespace),
            (r'(' + W + r'*)(?=//)', Whitespace, '#pop'),
            (r'([-+*/%])(\s*)', bygroups(BinOp, Whitespace), 'trailing_value'),
            (r'(\.)(\s*)', bygroups(BinOp, Whitespace), 'trailing_dotable_value'),
            (r'(' + W + r'*)(?=\()', Whitespace, 'func_expr'),
            default('#pop'),  # pop out when no operators are seen
        ],
        'trailing_value': [
            (r'(\s*)(//.*)(\s*)', bygroups(Whitespace, Comment.Single, Whitespace)),
            (r'\s+', Whitespace),
            include('value'),
        ],
        'trailing_dotable_value': [
            (r'(\s*)(//.*)(\s*)', bygroups(Whitespace, Comment.Single, Whitespace)),
            (r'\s+', Whitespace),
            include('dotable_value'),
        ],
        'dqs_key': dqs(DictKey),
        'sqs_key': sqs(DictKey),
        'dqs': dqs(String.Double),
        'sqs': sqs(String.Single),
        'tdqs': [
            (r'"""', TripleD, '#pop'),
            (r'\\\\|\\"|\\|"', TripleD),  # allow escapes
            (r'[^\\"]+', TripleD),  # not cater for multiple line
        ],
        'tsqs': [
            (r"'''", TripleS, '#pop'),
            (r"\\\\|\\'|\\|'", TripleS),  # allow escapes
            (r"[^\\']+", TripleS),  # not cater for multiple line
        ],
        'numbers': [
            (r'[-+]?(\d+\.\d*|\d*\.\d+)([eE][+-]?[0-9]+)?j?', Number.Float, ('#pop', 'operators')),
            (r'[-+]?\d+[eE][+-]?[0-9]+j?', Number.Float, ('#pop', 'operators')),
            (r'[-+]?0[0-7]+j?', Number.Oct, ('#pop', 'operators')),
            (r'[-+]?0[bB][01]+', Number.Bin, ('#pop', 'operators')),
            (r'[-+]?0[xX][a-fA-F0-9]+', Number.Hex, ('#pop', 'operators')),
            include('integers'),
        ],
        'integers': [
            (r'[-+]?\d+L', Number.Integer.Long, ('#pop', 'operators')),
            (r'[-+]?\d+j?', Number.Integer, ('#pop', 'operators')),
        ],
    }
    flags = re.MULTILINE

    def __init__(self, stack=None, **options):
        super().__init__(**options)
        self.stack = stack or ('root',)

    def get_tokens_unprocessed(self, text, stack=None) -> Iterable[PeekToken]:
        """
        Convert DictKey to common string if it is part of an expression
        """
        stack = stack or self.stack
        stream = super().get_tokens_unprocessed(text, stack)
        buffer = []
        while True:
            try:
                pt = PeekToken(*next(stream))
                if pt.ttype is DictKey:
                    assert len(buffer) == 0
                    buffer.append(pt)
                    while True:
                        pt_next = PeekToken(*next(stream))
                        if pt_next.ttype in (DictKey, Whitespace, Comment.Single):
                            buffer.append(pt_next)
                        elif pt_next.ttype is Colon:
                            buffer.append(pt_next)
                            for t in buffer:
                                yield t
                            buffer = []
                            break
                        else:
                            buffer.append(pt_next)
                            first_token = buffer[0]
                            if first_token.value == "'":
                                actual_type = String.Single
                            elif first_token.value == '"':
                                actual_type = String.Double
                            elif first_token.value == "'''":
                                actual_type = String.TripleS
                            elif first_token.value == '"""':
                                actual_type = String.TripleD
                            else:
                                raise PeekSyntaxError(text, first_token, message='DictKey expected')
                            for t in buffer:
                                if t.ttype is DictKey:
                                    t = PeekToken(t.index, actual_type, t.value)
                                yield t
                            buffer = []
                            break
                else:
                    yield pt

            except StopIteration:
                if buffer:
                    for t in buffer:
                        yield t
                break


Slash = Punctuation.Slash
PathPart = Text.PathPart
QuestionMark = Punctuation.QuestionMark
Ampersand = Punctuation.Ampersand
Pound = Punctuation.Pound
ParamName = Name.ParamName
ParamValue = Text.ParamValue


class UrlPathLexer(RegexLexer):
    name = 'URLPATH'
    aliases = ['url_path']

    tokens = {
        'root': [(r'/', Slash), (r'[^/?\s]+', PathPart), (r'\?', QuestionMark, 'query')],
        'query': [
            (r'&', Ampersand),
            (r'([^=&\s]+)(=)?([^&#\s]+)?', bygroups(ParamName, Assign, ParamValue)),
            (r'#', Pound, ('#pop', 'fragment')),
        ],
        'fragment': [
            (r'\S+', Text),
        ],
    }

    def get_tokens_unprocessed(self, text, stack=('root',)) -> PeekToken:
        for t in super().get_tokens_unprocessed(text, stack):
            yield PeekToken(*t)



"""Console script for peek."""
import argparse
import sys

from peek import __version__
from peek.config import config_location
from peek.peekapp import PeekApp


def main():
    """Console script for peek."""
    parser = argparse.ArgumentParser()

    parser.add_argument('input', nargs='*', help='script files')

    parser.add_argument('--config', default=config_location() + 'peekrc', help='Configuration file to load')

    parser.add_argument('-e', '--extra-config-option', action='append', help='Extra configuration option to override')

    parser.add_argument('--name', help='A friendly name for the connection')
    parser.add_argument('--hosts', default=argparse.SUPPRESS, help='ES hosts to connect to (default localhost:9200)')
    parser.add_argument('--cloud_id', help='Elastic Cloud ID')
    parser.add_argument('--username', help='Username')
    parser.add_argument('--password', help='Password')
    parser.add_argument('--api_key', help='API key of format id:key')
    parser.add_argument('--token', help='Token for authentication')
    parser.add_argument(
        '--use_ssl', action='store_true', default=argparse.SUPPRESS, help='Enable TLS for connecting to ES'
    )
    parser.add_argument(
        '--verify_certs', action='store_true', default=argparse.SUPPRESS, help='Verify server certificate'
    )
    parser.add_argument('--assert_hostname', default=argparse.SUPPRESS, help='Verify for the given hostname')
    parser.add_argument('--assert_fingerprint', default=argparse.SUPPRESS, help='Verify for the given fingerprint')
    parser.add_argument('--ca_certs', help='Location of CA certificates')
    parser.add_argument('--client_cert', help='Location of client certificate')
    parser.add_argument('--client_key', help='Location of client private key')
    parser.add_argument(
        '--force_prompt', action='store_true', default=argparse.SUPPRESS, help='Force prompting for password'
    )
    parser.add_argument(
        '--no_prompt', action='store_true', default=argparse.SUPPRESS, help='Do not prompt for password'
    )
    parser.add_argument('-z', '--zero_connection', action='store_true', help='Start the session with no connection')

    parser.add_argument('-V', '--version', action='version', version=__version__)

    ns = parser.parse_args()

    isatty = sys.stdin.isatty()
    batch_mode = (not isatty) or bool(ns.input)

    peek = PeekApp(
        batch_mode=batch_mode,
        config_file=ns.config,
        extra_config_options=ns.extra_config_option,
        cli_ns=ns,
    )
    if not batch_mode:
        peek.run()
    else:
        if ns.input:
            for f in ns.input:
                with open(f) as ins:
                    peek.process_input(ins.read())
        else:
            stdin_read = sys.stdin.read()
            peek.process_input(stdin_read)

    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover
